/* see LICENSE for license details */

#include <sys/types.h>
#include <sys/ioctl.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

#include "arg.h"

/* MACROS */
#define ABUF_INIT {NULL, 0}
#define CTRL_KEY(k) ((k) & 0x1f)


/* TYPES */
struct abuf {
	char *b;
	int len;
};

/* editor row: stores a row of text */
typedef struct eRow {
	char *chars;          /* text read from file */
	char *rchars;         /* text to render */
	int size;             /* length of chars */
	int rsize;            /* length of rchars */
	unsigned char *hl;    /* array of rsize storing highlight type per rchar */
} eRow;

struct editor_config {
	int cursor_x, cursor_y, cursor_rx;
	int screen_rows, screen_columns;
	int num_rows;
	eRow *row;
	int row_offset, column_offset;
	char *filename;
	struct termios orig_termios;
	int mode; /* 0: normal, 1: insert, 2: visual */
	int show_status;
	char status_msg[80];
	time_t status_msg_time;
	int file_changed;
	char *find_query;
	int find_direction;
};

/* color scheme types */
enum { ColorNormal, ColorDigit, ColorComment, ColorKeyword };
enum { FG, BG, STYLE };

/* configuration, allows nested code to access above variables */
#include "config.h"


/* FUNCTION DECLARATIONS */
char* editor_prompt(char *prompt);
void editor_set_status_message(const char *fmt, ...);


/* GLOBAL VARIABLES */
struct editor_config E;
char *argv0;

/* FUNCTION DEFINITIONS */

/** TERMINAL **/
void
die(const char *s)
{
	/* clear screen on exit */
	write(STDOUT_FILENO, "\x1b[2J", 4);
	write(STDOUT_FILENO, "\x1b[H", 3);
	perror(s);
	exit(1);
}

void
disable_raw_mode()
{
	if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &E.orig_termios) == -1)
		die("tcsetattr in disable_raw_mode()");
}

void
enable_raw_mode()
{
	struct termios raw;

	if (tcgetattr(STDIN_FILENO, &E.orig_termios) == -1)
		die("tcgetattr in enable_raw_mode()");
	atexit(disable_raw_mode);

	/* fix modifier keys, set 8 bits per char, ignore interrupts */
	raw = E.orig_termios;
	raw.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
	raw.c_oflag &= ~(OPOST);
	raw.c_cflag |= (CS8);
	raw.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);

	/* set read() timeout in tenths of seconds */
	raw.c_cc[VMIN] = 0;
	raw.c_cc[VTIME] = 1;

	/* apply terminal settings */
	if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &raw) == -1)
		die("tcsetattr in enable_raw_mode()");
}

char
editor_read_key()
{
	int nread;
	char c;
	while ((nread = read(STDIN_FILENO, &c, 1)) != 1) {
		if (nread == -1 && errno != EAGAIN)
			die("read in editor_read_key()");
	}
	return c;
}

/* get screen size by moving cursor and get current position */
int
get_cursor_position(int *rows, int *cols) {
	char buf[32];
	unsigned int i;

	if (write(STDOUT_FILENO, "\x1b[6n", 4) != 4)
		return -1;

	i = 0;
	while (i < sizeof(buf) - 1) {
		if (read(STDIN_FILENO, &buf[i], 1) != 1)
			break;
		if (buf[i] == 'R')
			break;
		++i;
	}
	buf[i] = '\0';

	if (buf[0] != '\x1b' || buf[1] != '[')
		return -1;
	if (sscanf(&buf[2], "%d;%d", rows, cols) != 2)
		return -1;
	return 0;
}

int
get_window_size(int *rows, int *cols)
{
	struct winsize ws;

	if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) == -1 || ws.ws_col == 0) {
		/* fallback screen size detection */
		if (write(STDOUT_FILENO, "\x1b[999C\x1b[999B", 12) != 12)
			return -1;
		return get_cursor_position(rows, cols);
	} else {
		*cols = ws.ws_col;
		*rows = ws.ws_row;
		return 0;
	}
}


/** SYNTAX HIGHLIGHTING **/

void
editor_update_syntax(eRow *row)
{
	int i;
	row->hl = realloc(row->hl, row->rsize);
	memset(row->hl, ColorNormal, row->rsize);
	for (i=0; i<row->rsize; ++i) {
		if (isdigit(row->rchars[i]))
			row->hl[i] = ColorDigit;
	}
}


/** ROW OPERATIONS **/

/* navigate over tab-representative string as one character */
int
editor_row_cursor_x_to_rx(eRow *row, int cursor_x)
{
	int rx, j;
	rx = 0;
	for (j=0; j<cursor_x; ++j) {
		if (row->chars[j] == '\t')
			rx += (tab_width - 1) - (rx % tab_width);
		rx++;
	}
	return rx;
}

/* translate on-screen position to data position */
int
editor_row_cursor_rx_to_x(eRow *row, int cursor_rx)
{
	int cur_rx, cx;
	cur_rx = 0;
	for (cx=0; cx<row->size; ++cx) {
		if (row->chars[cx] == '\t')
			cur_rx += (tab_width - 1) - (cur_rx % tab_width);
		cur_rx++;

		if (cur_rx > cursor_rx)
			return cx;
	}
	return cx;
}

/* translate tabs before display */
void
editor_row_update(eRow* row)
{
	int j, idx, tabs;

	free(row->rchars);
	row->rchars = malloc(row->size + 1);

	tabs = 0;
	for (j=0; j<row->size; ++j)
		if (row->chars[j] == '\t')
			tabs++;

	free(row->rchars);
	row->rchars = malloc(row->size + tabs*(tab_width - 1) + 1);

	idx = 0;
	for (j=0; j<row->size; ++j) {
		if (row->chars[j] == '\t') {
			row->rchars[idx++] = '>';
			while (idx % tab_width != 0)
				row->rchars[idx++] = ' ';
		} else {
			row->rchars[idx++] = row->chars[j];
		}
	}
	row->rchars[idx] = '\0';
	row->rsize = idx;
	editor_update_syntax(row);
}

/* add row to buffer */
void
editor_row_insert(int i, char *s, size_t len)
{
	if (i<0 || i>E.num_rows)
		return;

	E.row = realloc(E.row, sizeof(eRow) * (E.num_rows + 1));
	memmove(&E.row[i+1], &E.row[i], sizeof(eRow) * (E.num_rows - i));

	E.row[i].size = len;
	E.row[i].chars = malloc(len + 1);
	memcpy(E.row[i].chars, s, len);
	E.row[i].chars[len] = '\0';

	E.row[i].rsize = 0;
	E.row[i].rchars = NULL;
	E.row[i].hl = NULL;
	editor_row_update(&E.row[i]);

	++E.num_rows;
	E.file_changed = 1;
}

/* insert character to row, making sure to allocate memory accordingly */
void
editor_row_insert_char(eRow *row, int i, int c)
{
	if (i<0 || i>row->size)
		i = row->size;
	row->chars = realloc(row->chars, row->size + 2);
	memmove(&row->chars[i+1], &row->chars[i], row->size - i+1);
	row->size++;
	row->chars[i] = c;
	editor_row_update(row);
	E.file_changed = 1;
}

/* append a string to the end of a row */
void
editor_row_append_string(eRow *row, char *s, size_t len)
{
	row->chars = realloc(row->chars, row->size + len + 1);
	memcpy(&row->chars[row->size], s, len);
	row->size += len;
	row->chars[row->size] = '\0';
	editor_row_update(row);
	E.file_changed = 1;
}

void
editor_row_delete_char(eRow *row, int i)
{
	if (i<0 || i>=row->size)
		return;
	memmove(&row->chars[i], &row->chars[i+1], row->size - i);
	row->size--;
	editor_row_update(row);
	E.file_changed = 1;
}

void
editor_row_free(eRow *row)
{
	free(row->chars);
	free(row->rchars);
	free(row->hl);
}

void
editor_row_delete(int i)
{
	if (i<0 || i>=E.num_rows)
		return;
	editor_row_free(&E.row[i]);
	memmove(&E.row[i], &E.row[i+1], sizeof(eRow)*(E.num_rows - i - 1));
	E.num_rows--;
	E.file_changed = 1;
}

void
editor_insert_char(int c)
{
	if (E.cursor_y == E.num_rows)
		editor_row_insert(E.num_rows, "", 0);
	editor_row_insert_char(&E.row[E.cursor_y], E.cursor_x, c);
	E.cursor_x++;
}

void
editor_insert_new_line()
{
	eRow *row;
	if (E.cursor_x == 0) {
		editor_row_insert(E.cursor_y, "", 0);
	} else {
		row = &E.row[E.cursor_y];
		editor_row_insert(E.cursor_y + 1, &row->chars[E.cursor_x],
		                  row->size - E.cursor_x);
		row = &E.row[E.cursor_y];
		row->size = E.cursor_x;
		row->chars[row->size] = '\0';
		editor_row_update(row);
	}
	E.cursor_y++;
	E.cursor_x = 0;
}

/* delete a character before the cursor, and allow for deletion across rows */
void
editor_delete_char_left()
{
	eRow *row;
	if (E.cursor_y == E.num_rows || (E.cursor_x == 0 && E.cursor_y == 0))
		return;
	
	row = &E.row[E.cursor_y];
	if (E.cursor_x > 0) {
		editor_row_delete_char(row, E.cursor_x - 1);
		E.cursor_x--;
	} else {
		E.cursor_x = E.row[E.cursor_y - 1].size;
		editor_row_append_string(&E.row[E.cursor_y - 1],
		                         row->chars, row->size);
		editor_row_delete(E.cursor_y);
		E.cursor_y--;
	}
}

void
editor_delete_char_right()
{
	if (E.cursor_y == E.num_rows)
		return;
	eRow *row = &E.row[E.cursor_y];
	editor_row_delete_char(row, E.cursor_x);
}

/* convert rows to one long char array of length buflen */
char*
editor_concatenate_rows(int *buflen)
{
	int totlen, j;
	char *buf, *p;

	totlen = 0;
	for (j=0; j<E.num_rows; ++j)
		totlen += E.row[j].size + 1;  /* add space for newline char */
	*buflen = totlen;

	buf = malloc(totlen);
	p = buf;
	for (j=0; j<E.num_rows; ++j) {
		memcpy(p, E.row[j].chars, E.row[j].size);
		p += E.row[j].size;
		*p = '\n';
		p++;
	}
	return buf;
}


/** FILE IO **/

/* open file and read lines into memory */
void
file_open(char *filename)
{
	free(E.filename);
	E.filename = strdup(filename);

	FILE *fp;
	char *line;
	size_t linecap;
	ssize_t linelen;
	
	fp = fopen(filename, "r");
	if (!fp)
		die("fopen in file_open");

	line = NULL;
	linecap = 0;
	while ((linelen = getline(&line, &linecap, fp)) != -1) {
		while (linelen > 0 && (line[linelen - 1] == '\n' ||
		                       line[linelen - 1] == '\r'))
			linelen--;
		editor_row_insert(E.num_rows, line, linelen);
	}
	free(line);
	fclose(fp);
	E.file_changed = 0;
}

void
file_save(char *filename)
{
	int len, fd;
	char *buf;

	if (filename == NULL) {
		filename = editor_prompt("save as: %s");
		if (filename == NULL) {
			editor_set_status_message("save aborted");
			return;
		}
	}

	buf = editor_concatenate_rows(&len);

	fd = open(filename, O_RDWR | O_CREAT, 0644);
	if (fd != -1) {
		if (ftruncate(fd, len) != -1) {
			if (write(fd, buf, len) == len) {
				close(fd);
				free(buf);
				E.filename = filename;
				E.file_changed = 0;
				editor_set_status_message("%d bytes written to disk", len);
				return;
			}
		}
		close(fd);
	}
	free(buf);
	editor_set_status_message("error: can't save! I/O error %s",
	                          strerror(errno));
}


/** FIND **/

/* reverse of strstr (3) */
char*
strrstr(char *haystack, char *needle,
        size_t haystack_length, size_t needle_length)
{
  char *cp;
  for (cp = haystack + haystack_length - needle_length;
       cp >= haystack;
       cp--) {
    if (strncmp(cp, needle, needle_length) == 0)
        return cp;
  }
  return NULL;
} 

/* find E.find_query from current cursor position moving in E.direction 
 * if opposite_direction = 0, and opposite E.direction if
 * opposite_direction = 1 */
void
editor_find_next_occurence(int opposite_direction)
{
	int y, y_inc, x_offset, query_len;
	eRow *row;
	char *match;

	if (!E.find_query)
		return;

	if ((E.find_direction && !opposite_direction) ||
		(!E.find_direction && opposite_direction))
		y_inc = +1;
	else
		y_inc = -1;

	x_offset = 0;
	query_len = strlen(E.find_query);
	/* from cursor until end of document */
	for (y = E.cursor_y;
	     y != ((y_inc == +1) ? E.num_rows : -1);
	     y += y_inc) {
		row = &E.row[y];

		if (y == E.cursor_y)
			x_offset = y_inc + E.cursor_x;
		else
			x_offset = 0;

		if (y_inc == +1) {
			match = strstr(row->chars + x_offset, E.find_query);
			if (match)
				break;
		} else {
			match = strrstr(row->chars, E.find_query,
			                (y == E.cursor_y) ? E.cursor_x : row->size,
			                query_len);
			if (match)
				break;
		}
	}

	if (match) {
		E.cursor_y = y;
		E.cursor_x = match - row->chars;
		/* E.row_offset = E.num_rows; */ /* put line to top of screen */

	} else {
		/* from other end of file until cursor */
		for (y = (y_inc == +1) ? 0 : E.num_rows - 1;
		     y != E.cursor_y + y_inc;
		     y += y_inc) {
			row = &E.row[y];
			match = strstr(row->chars, E.find_query);
			if (match)
				break;
		}
		if (match) {
			E.cursor_y = y;
			E.cursor_x = editor_row_cursor_rx_to_x(row, match - row->chars);
		}
	}
}

/* direction = 1 is forward, 0 is backward */
void
editor_find(int direction)
{
	char *query;

	if (direction)
		query = editor_prompt("/%s");
	else
		query = editor_prompt("?%s");
	if (query == NULL)
		return;

	E.find_direction = direction;
	free(E.find_query);
	E.find_query = strdup(query);
	editor_find_next_occurence(0);
	free(query);
}


/** OUTPUT **/

void
editor_update_screen_size()
{
	if (get_window_size(&E.screen_rows, &E.screen_columns) == -1)
		die("get_window_size in main()");
	E.screen_rows--;
	if (E.show_status)
		E.screen_rows--;
}

/* reallocate append buffer to hold string s */
void
ab_append(struct abuf *ab, const char *s, int len)
{
	char *new;
	new = realloc(ab->b, ab->len + len);

	if (new == NULL)
		return;
	memcpy(&new[ab->len], s, len);
	ab->b = new;
	ab->len += len;
}

void
ab_free(struct abuf *ab) {
	free(ab->b);
}

int
show_status_message()
{
	if (E.show_status &&
	    strlen(E.status_msg) &&
	    time(NULL) - E.status_msg_time < status_message_timeout) {
		return 1;
	} else {
		if (E.show_status) {
			E.show_status = 0;
			E.screen_rows++;
		}
		return 0;
	}
}

/* draw status line consisting of left and right components.
 * if the screen is narrower than both parts, just show the left status, 
 * truncated, if necessary. */
void
editor_draw_status(struct abuf *ab)
{
	char left_status[512], right_status[512];
	int left_status_len, right_status_len, padding;
	int percentage;

	left_status_len = 0;
	switch (E.mode) {
		case 1:
			left_status_len = snprintf(left_status, sizeof(left_status),
			                           "INSERT ");
			break;
		case 2:
			left_status_len = snprintf(left_status, sizeof(left_status),
			                           "VISUAL ");
			break;
	}
	left_status_len += snprintf(left_status + left_status_len,
	                            sizeof(left_status),
	                            "%.20s %s",
	                            E.filename ? E.filename : "[unnamed]",
	                            E.file_changed ? "[+] " : "");

	percentage = (int)((float)(E.cursor_y)/(E.num_rows-1)*100);
	if (percentage < 0) percentage = 0;
	right_status_len = snprintf(right_status, sizeof(right_status),
	                            "%d%% < %d, %d",
	                            percentage,
	                            E.cursor_x+1, E.cursor_y+1);

	if (left_status_len > E.screen_columns)
		left_status_len = E.screen_columns;

	padding = E.screen_columns - left_status_len - right_status_len;
	if (padding < 0) {
		if (left_status_len < E.screen_columns)
			ab_append(ab, left_status, left_status_len);
		else
			ab_append(ab, left_status, E.screen_columns);
	} else {
		ab_append(ab, left_status, left_status_len);
		while (padding--)
			ab_append(ab, " ", 1);
		ab_append(ab, right_status, right_status_len);
	}

	if (show_status_message()) {
		ab_append(ab, "\x1b[m", 3);
		ab_append(ab, "\r\n", 2);
	}
}

/* draw status message if as long as it is specified and it is not too old */
void
editor_draw_status_message(struct abuf *ab)
{
	int msglen;
	ab_append(ab, "\x1b[K", 3);
	msglen = strlen(E.status_msg);
	if (msglen > E.screen_columns)
		msglen = E.screen_columns;
	if (show_status_message())
		ab_append(ab, E.status_msg, msglen);
}

/* set vertical offset between file and screen when hitting the boundaries */
void
editor_scroll()
{
	E.cursor_rx = 0;
	if (E.cursor_y < E.num_rows)
		E.cursor_rx = editor_row_cursor_x_to_rx(&E.row[E.cursor_y],
		                                        E.cursor_x);

	if (E.cursor_y < E.row_offset)
		E.row_offset = E.cursor_y;
	else if (E.cursor_y >= E.row_offset + E.screen_rows)
		E.row_offset = E.cursor_y - E.screen_rows + 1;

	if (E.cursor_rx < E.column_offset)
		E.column_offset = E.cursor_rx;
	else if (E.cursor_rx >= E.column_offset + E.screen_columns)
		E.column_offset = E.cursor_rx - E.screen_columns + 1;
}

/* draw editor screen.
 * show tilde characters after EOF, and show status on last line */
void
editor_draw_rows(struct abuf *ab)
{
	int y, j, len, file_row;
	char *c;
	unsigned char *hl;
	char buf[16];
	for (y = 0; y < E.screen_rows; ++y) {
		file_row = y + E.row_offset;
		if (file_row < E.num_rows) {
			len = E.row[file_row].rsize - E.column_offset;
			if (len < 0)
				len = 0;
			if (len > E.screen_columns)
				len = E.screen_columns;
			c = &E.row[file_row].rchars[E.column_offset];
			hl = &E.row[file_row].hl[E.column_offset];
			for (j = 0; j < len; ++j) {
				if (hl[j] == ColorDigit) {
					snprintf(buf, sizeof(buf),
					         "\x1b[3%dm", colors[ColorDigit][FG]); 
					ab_append(ab, buf, strlen(buf));
					/* ab_append(ab, "\x1b[31m", 5); */
					ab_append(ab, &c[j], 1);
					ab_append(ab, "\x1b[39m", 5);
				} else {
					ab_append(ab, &c[j], 1);
				}
			}
		} else {
			ab_append(ab, "~", 1);
		}

		ab_append(ab, "\x1b[K", 3); /* erase to end of line */
		ab_append(ab, "\r\n", 2);
	}
}

/* fill output append buffer and write to terminal.
 * move cursor to left before repaint, and to cursor position after.
 * hide the cursor when repainting.
 * VT100 escape sequences:
 * - http://vt100.net/docs/vt100-ug/chapter3.html
 * - http://vt100.net/docs/vt100-ug/chapter3.html#CUP */
void
editor_refresh_screen()
{
	char buf[32];
	struct abuf ab = ABUF_INIT;

	show_status_message();
	editor_scroll();

	ab_append(&ab, "\x1b[?25l", 6); /* hide cursor */
	ab_append(&ab, "\x1b[H", 3);    /* cursor to home */
	
	editor_draw_rows(&ab);
	editor_draw_status(&ab);
	editor_draw_status_message(&ab);

	snprintf(buf, sizeof(buf), "\x1b[%d;%dH",
	        (E.cursor_y - E.row_offset)+1,
	        (E.cursor_rx - E.column_offset)+1);
	ab_append(&ab, buf, strlen(buf));

	ab_append(&ab, "\x1b[?25h", 6); /* show cursor */

	write(STDOUT_FILENO, ab.b, ab.len);
	ab_free(&ab);
}

void
editor_place_cursor(int cx, int cy)
{
	char buf[128];
	snprintf(buf, sizeof(buf), "\x1b[%d;%dH", cy, cx);
	write(STDOUT_FILENO, buf, strlen(buf));
}

/* set status message text, uses same format as printf */
void
editor_set_status_message(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vsnprintf(E.status_msg, sizeof(E.status_msg), fmt, ap);
	va_end(ap);
	E.status_msg_time = time(NULL);

	if (!E.show_status) {
		E.screen_rows--;
		E.show_status = 1;
	}
}


/** INPUT **/

/* prompt is expected to be a format string containing a %s */
char*
editor_prompt(char *prompt)
{
	size_t bufsize, buflen;
	char *buf;
	int c;
	
	bufsize = 128;
	buflen = 0;
	buf = malloc(bufsize);
	buf[0] = '\0';

	while (1) {
		editor_update_screen_size();
		editor_set_status_message(prompt, buf);
		editor_refresh_screen();
		editor_place_cursor(strlen(prompt) - 1 + strlen(buf),
		                    E.screen_rows + 2);

		c = editor_read_key();
		if (c == CTRL_KEY('h') || c == 127) { /* detect backspace */
			if (buflen != 0)
				buf[--buflen] = '\0';
		} else if (c == '\x1b' || (c == '\r' && !buflen)) { /* detect escape */
			editor_set_status_message("");
			free(buf);
			return NULL;
		} else if (c == '\r') {
			if (buflen != 0) {
				editor_set_status_message("");
				return buf;
			}
		} else if (!iscntrl(c) && c < 128) {
			if (buflen >= bufsize - 1) {
				bufsize *= 2;
				buf = realloc(buf, bufsize);
			}
			buf[buflen++] = c;
			buf[buflen] = '\0';
		}
	}
}

/* move cursor according to screen, file, and line limits */
void
editor_move_cursor(char key)
{
	int row_len;
	eRow *row;
	
	switch(key) {
		case 'h':
			if (E.cursor_x != 0) {
				E.cursor_x--;
			} else if (E.cursor_y > 0) {
				E.cursor_y--;
				E.cursor_x = E.row[E.cursor_y].size;
			}
			break;
		case 'j':
			if (E.cursor_y < E.num_rows - 1)
				E.cursor_y++;
			break;
		case 'k':
			if (E.cursor_y != 0)
				E.cursor_y--;
			break;
		case 'l':
			row = (E.cursor_y >= E.num_rows) ? NULL : &E.row[E.cursor_y];
			if (row && E.cursor_x < row->size - 1) {
				E.cursor_x++;
			} else if (row && E.cursor_x == row->size &&
			           E.cursor_y < E.num_rows - 1) {
				E.cursor_y++;
				E.cursor_x = 0;
			}
			break;
	}

	/* do not allow navigation past EOL by vertical navigation */
	row = (E.cursor_y >= E.num_rows) ? NULL : &E.row[E.cursor_y];
	row_len = row ? row->size : 0;
	if (E.cursor_x > row_len)
		E.cursor_x = row_len;
}

void
editor_read_command()
{
	char* query;
	query = editor_prompt(":%s");

	editor_set_status_message("you typed: '%s'", query);
	free(query);
}

/* first word in query is the command, the remaining are interpreted as args */
void
editor_command(char *query)
{
	switch(query) {

	}
}


void
editor_process_keypress()
{
	char c;
	int i;
	
	c = editor_read_key();

	if (E.mode == 0) {  /* normal mode */
		switch (c) {
			case 'i':
				E.mode = 1;
				break;
			case 'a':
				editor_move_cursor('l');
				E.mode = 1;
				break;

			case 'h':
			case 'j':
			case 'k':
			case 'l':
				editor_move_cursor(c);
				break;

			case ':':
				editor_read_command();
				break;

			case LEADER:
				c = editor_read_key();
				switch (c) {
					case 'w':
						file_save(E.filename);
						break;
					case 'q':
						if (E.file_changed) {
							editor_set_status_message(
								"error: file has unsaved changes! "
								"Press :q! to confirm quit");
							break;
						} else {
							/* clear screen */
							write(STDOUT_FILENO, "\x1b[2J", 4); 
							write(STDOUT_FILENO, "\x1b[H", 3);
							exit(0);
							break;
						}
				}
				break;

			case CTRL_KEY('f'):
				i = E.screen_rows;
				while (i--)
					editor_move_cursor('j');
				break;
			case CTRL_KEY('b'):
				i = E.screen_rows;
				while (i--)
					editor_move_cursor('k');
				break;

			case CTRL_KEY('d'):
				i = E.screen_rows/2;
				while (i--)
					editor_move_cursor('j');
				break;
			case CTRL_KEY('u'):
				i = E.screen_rows/2;
				while (i--)
					editor_move_cursor('k');
				break;

			case '0':
				E.cursor_x = 0;
				break;
			case '$':
				if (E.cursor_y < E.num_rows)
					E.cursor_x = E.row[E.cursor_y].size - 1;
				break;

			case 'g':
				c = editor_read_key();
				if (c == 'g') {
					E.cursor_x = 0;
					E.cursor_y = 0;
				}
				break;
			case 'G':
				E.cursor_x = 0;
				E.cursor_y = E.num_rows - 1;
				break;

			case 'x':
				editor_delete_char_right();
				break;
			case 'd':
				c = editor_read_key();
				if (c == 'd') {
					editor_row_delete(E.cursor_y);
					editor_move_cursor('h');
				}
				break;

			case 'o':
				if (E.cursor_y < E.num_rows)
					E.cursor_x = E.row[E.cursor_y].size;
				editor_insert_new_line();
				E.mode = 1;
				break;
			case 'O':
				E.cursor_x = 0;
				editor_insert_new_line();
				editor_move_cursor('k');
				E.mode = 1;
				break;

			case 'I':
				E.cursor_x = 0;
				E.mode = 1;
				break;
			case 'A':
				if (E.cursor_y < E.num_rows)
					E.cursor_x = E.row[E.cursor_y].size;
				E.mode = 1;
				break;

			case '/':
				editor_find(1);
				break;
			case '?':
				editor_find(0);
				break;
			case 'n':
				editor_find_next_occurence(0);
				break;
			case 'N':
				editor_find_next_occurence(1);
				break;
		}
	} else if (E.mode == 1) {  /* insert mode */
		switch (c) {
			case CTRL_KEY('c'):
			case '\x1b':  /* escape */
				E.mode = 0;
				break;
			
			case CTRL_KEY('\r'): /* enter */
				editor_insert_new_line();
				break;

			case 127:  /* backspace */
			case CTRL_KEY('h'):
				editor_delete_char_left();
				break;

			case CTRL_KEY('l'):
				break;

			default:
				editor_insert_char(c);
				break;
		}
	}
}


/** INIT **/

void
deinit_editor() {
	int i;
	free(E.filename);
	free(E.find_query);
	for (i=0; i<E.num_rows; ++i)
		editor_row_free(&E.row[i]);
	free(E.row);
}

/* set editor state variables, make room for status */
void
init_editor()
{
	E.cursor_x = 0;
	E.cursor_y = 0;
	E.cursor_rx = 0;
	E.mode = 0;
	E.num_rows = 0;
	atexit(deinit_editor);
	E.row = NULL;
	E.row_offset = 0;
	E.column_offset = 0;
	E.filename = NULL;
	E.status_msg[0] = '\0';
	E.status_msg_time = 0;
	E.show_status = 0;
	E.file_changed = 0;
	E.find_query = NULL;
}

void
usage()
{
	printf("%s: %s [OPTIONS] [FILES]\n"
	       "\nOptional arguments:\n"
	       " -h        show this message\n"
	       " -v        show version information\n",
	       __func__, PROGNAME);
}

void
version()
{
	printf("%s version %s\n"
	       "Licensed under the GNU Public License, v3+\n"
	       "written by Anders Damsgaard, anders@adamsgaard.dk\n"
	       "https://gitlab.com/admesg/ve\n",
	       PROGNAME, VERSION);
}

/** MAIN **/
int
main(int argc, char *argv[])
{
	ARGBEGIN {
	case 'h':
		usage();
		return 0;
	case 'v':
		version();
		return 0;
	default:
		usage();
	} ARGEND;

	enable_raw_mode();
	init_editor();

	editor_set_status_message("%s %s", PROGNAME, VERSION);

	if (argv[0])
		file_open(argv[0]);

	while (1) {
		editor_update_screen_size();
		editor_refresh_screen();
		editor_process_keypress();
	}
	return 0;
}
